// ================================================================
// Morse Practice   https://www.instructables.com/id/Morse-Code-Trainer/
// Currently using: STEM / Educational Game Platform - using a top level Menu selection
// with a Micro Controller, 8 LED 4 color bar and 4 Button module + piezo sounder
//
// license: Attribution-NonCommercial-ShareAlike
// Aug-Sep 2020 Ron Miller, Initial Development
// Jan & Aug 2024 Updated RRM

// use one of the next set of include files
#include "CW_Words.h"      // 900 of the most common words + HAM&Tech terms ...
// following vocabulary sets have a better balance of the letters A-Z for more even practice
//#include "CW_set2_Words.h"  // 900 words from top 500-1500 (=# wrds starting with each of a-z) + HAM...
//#include "CW_AZS_Words.h"  // words from A-Z Sentences (Pangrams) any 10 words may have all 26 letters ...

int numWrds=sizeof(words)/sizeof(void *); // Nano size of an pointer is 2 (but 4 for RP2040)
char wrd[21];
  
//  include Library of Hardware Support functions
#include "4Btn_8LED_Platform.h"

#include <EEPROM.h>
// the address in the EEPROM we're going to write to
int eeSaveAddr = 73;  // arbitary, HAM for Best to you
unsigned int savedPeriod, savedFF;


bool spkr = true; // speakers & pasive piezos can be AC driven for more volume (not active buzzers)
bool quiet = false; // assumes a 150 ohm resister between AUDIO_NEG & GND

#define DEBUG 0
#if DEBUG
  #define debug_(x) Serial.print(x)
  #define debug_ln(x) Serial.println(x)
#else
  #define debug_(x)
  #define debug_ln(x)
#endif

bool debug = DEBUG;  //#define debug false

#define info_(x) // Serial.print(x)
#define info_ln(x) // Serial.println(x)

void Nada() {;}
//#define debugDit printDit
#define debugDit(x) Nada()

// ------------------------ function prototypes
void freeForm();
void manageRate();
void MC_RxTx_Practice();
void Adv_Relay_exercise(bool);
void Tx_Line_test();

byte fetchMorseChar();
char dida2Ascii(byte);
byte Ascii2dida(char);
void set_didaStr(byte, char[]);

void dit_dah(unsigned short int );
void ditAudio(unsigned short int);
void dahAudio(unsigned short int);
void playMorseCode(byte, int);
void playStr(char*, int);

void showLevel(byte);
byte get_Mode(byte);
byte get_Option(byte, int);
bool YesNo();
void ranIndexes(byte[], byte);
long mcRNum;  // random seed for Shuffled word select sequence

int wInx=0;  // global MC word Index, to keep the words used extra fresh & new
unsigned int pinx; // global char selection reference inx
bool btnDown();
void stableBtns();
bool isPunct(byte);
#define IsPunct  (cinput>32 && !(toupper(cinput)>64 && toupper(cinput)<91)

unsigned long t0,t1;
byte func;

unsigned int avgPeriod, pAveraging;
unsigned int avgDit=800,avgDah=2400;
unsigned int ditPeriod;
int farns, deFF;  // Farnsworth Factor, multiplier for std to Farnsworth inter char & word periods
bool pOpen=true;  // expect an open ( else next should be close )
bool postAscii;  // print ascii after each character ditdah
bool enbBS=false;     // enables use of BS 0x08 output to monitor
bool echoChar;
bool easy=false;  // go easy on beginners, used during freeForm() 

bool oneKey=true; // straight single contact key   (TBD: false=using a 2 contact paddle)
//bool batOp=false; // would-be indicator for battery Operation (w/o serial monitor)
char didaStr[10];
unsigned long randSeed=0;
const char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";

const char nonAlpha[] = " 1234567890.,?/'-:\"+!=@$)&;\n(~";
                          //       1    2    3    4    5    6    7    8    9    0
const byte nonAlpha_dida[] = {0,0x10,0x18,0x1C,0x1E,0x1F,0xEF,0xE7,0xE3,0xE1,0xE0,
       0x2A,0xCC,0x33,0xED,0x21,0xDE,0xC7,0x2D,0x15,0xD4,0xEE,0x25,0x76,0xD2,0x17,0xD5,0x0A,0xE9};
     // .    ,    ?    /    '    -    :    "    +    !    =    @    $    )    &    ;   \n    (
  
// the setup routine runs once on power up or when you press reset:
//=================================================================
void setup() {

  // ------------ setup the Mini Game/STEM platform (8LEDs, 4Btns, and Audio)
  gPlatform_setup();

  // initialize serial communication
  Serial.begin(115200);
  info_ln("\n\n  * Morse CW Practice *");

  // ------ setup optional behavior based on button states at startup
  stableBtns(); // get btns state after debouncing
  if (Btn) {
    // WAS: ahead btn1 held at startup disables Farnsworth
    quiet = btn1; btn1-false; // now: enable quiet opperation
    if (btn2) oneKey = false;  // btn2 for 2-Key, single contact straight key (1 button) by default
    if (btn3) randomSeed(randSeed=314159); // always use the same rand #s for each run
    debug = btn4;
    // note btn4 toggles Loud during Main menu
    info_ln("button state at startup: "); 
    info_(btn1); info_(btn2); info_(btn3); info_ln(btn4);
  }
  if (!oneKey) Serial.println("2-Key operstion");
  if (quiet) pinMode(AUDIO_NEG, INPUT);
  
 //  ditPeriod= (60000)/(240*2);  // in msecs, for ~10WPM to start off  (240 in 60,000 ms is 5WPM)
  ditPeriod= (60000)/(240*3);  // in msecs, for ~15WPM to start off  (240 in 60,000 ms is 5WPM)
 //  ditPeriod= (60000)/(240*3+120);    // in msecs, for ~17.5WPM to start off 
 //  ditPeriod= (60000)/(240*4);  // in msecs, for ~20WPM to start off  (240 in 60,000 ms is 5WPM)
  
  // Farnsworth calculations
  int cWPM = 18; // Char-WPM speed
  int sWPM = 10;  // Spacing-WPM,    8: x4 farns  10: x3 farns, normally set to =10
  //btw:  farns with cWPM/sWPM of 10/18 timing result in ~ 16.5 WPM overall  (8/18 ~ 16)
  float fp = 1000. * (60./sWPM - 37.2/cWPM) / 19.; // inter time period of Farnsworth timing
  farns = fp/(1000*1.2/cWPM);  // factor to convert std to Farnsworth unit time period
  deFF=farns;

 /***/
  // read previously saved data from EEPROM
  savedPeriod = EEPROM.read(eeSaveAddr);
  savedFF = EEPROM.read(eeSaveAddr+sizeof(savedPeriod));
  info_(" savedPeriod: "); info_(savedPeriod);
  info_("ms   savedFF xfactor: "); info_ln(savedFF);
 /***/

  if (0<savedFF && savedFF<6 && 20<savedPeriod && savedPeriod<250) {
    ditPeriod = savedPeriod;
    deFF = farns = savedFF;
  }
  avgPeriod= 2*ditPeriod;
  pAveraging= 4*avgPeriod;

  if (btn1) {
    farns = 1; // disable Farnsworth effect if btn1 held down at power up.
 ///    Serial.println("Standard inter-char & word timing used.");
  } else {
 ///    Serial.print("Farnsworth ratio: "); Serial.print(sWPM); Serial.print("/"); Serial.print(cWPM);
 //    if (DEBUG) {Serial.print("   unit period factor: "); Serial.print(farns);}
    Serial.println();
  }

  Serial.print(" Keying rate (in WPM): "); Serial.print((1000.0*60.0) / (240.0*ditPeriod/5.0));
  Serial.print("  + Farnsworth spacing: x"); Serial.println(farns);

  delay(200);
  //  demonstrate there is Life in the components (LEDs)
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,1); set_LEDs(0x80>>i,1);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,2); set_LEDs(0x80>>i,2);}
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,3); set_LEDs(0x80>>i,3);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,0); set_LEDs(0x80>>i,0);}
  refreshWait(500);
 //  clearDisp();
  if (Serial) { // True if a Serial 'monitor' conection
    //Serial.print(" terminal string:");
    Serial.write(5); // send an ENQ enquiry
    delay(10);
    //Serial.println(Serial.available());
    if (Serial.available()!=0) enbBS=true;
  }
  // highlight the Audio
  boop(40,250);
  beep(40,250);
  buZZ(40,250);

 //  delay(1000);
}

void heartBeat() { // activate LEDs and make noise to show that #5 is alive
    for (int i=0; i<20*5; i++) {
      set_LEDs(i,2);
      beep(30,30);
      refreshWait(200);
    }
}

// the loop routine runs over and over again forever:
//=================================================================
void loop() { // Top Menu level, for selecting 1 of N activities
  byte maxFunc;

///  Serial.print(" Vocabuary of Words[] is "); Serial.println(numWrds);
  while(Btn) scanBtns();  // clear btns
  func=1;  // initial selection

  while(1) {
    clearDisp();
    while(Btn) {buZZ(1); scanBtns();}  // wait for no button

    maxFunc=7;  // currently there are 6 activities/games supported
    // indicate selectable functions 1-6
    for (int i=1; i<=maxFunc; i++) {dim[i]=true;}
    brightOne=func;
    refreshWait(20);
    
    // watch for btn presses other than btn1
    while (! btn1) {
      refreshDisp();
      delay(1);
      scanBtns();
      //if (btn2Pressed) brightOne = (brightOne==1)?  maxFunc : brightOne-1;  // -- move left
      if (btn2Pressed) brightOne = (brightOne==0)?  maxFunc : brightOne-1;  // -- move left Allow 0:debugging
      if (btn3Pressed) brightOne = (brightOne%maxFunc) + 1; // ++ moves right
      //if (btn4Pressed) batOp=true;
      // --------------- cycle: normal-quiet-Loud volume
      if (btn4Pressed) {
        if (Loud) { Loud=false; }
        else if (quiet) {quiet=false; Loud=true;}
        else quiet=true;
      }
      if (Loud && !spkr) {Loud=false; quiet=true;}
      if (quiet) pinMode(AUDIO_NEG, INPUT);
      else pinMode(AUDIO_NEG, OUTPUT);
      while (btn4) { dit_dah(20); scanBtns();}
      //----------------
//while (btn2) {buZZ(50,0); scanBtns();}
//while (btn3) {LOUDbuZZ(50); scanBtns();}
    }

    // randomize driven by real-world user event
    if (randSeed==0) {
      randomSeed(randSeed = micros()/4);  // micros() resolution is ~ 4us
      // used for QA testing: randomSeed(randSeed=36256276);  // starts with Rx test Ln=8 (7 w/ opt<3)
//     Serial.print("random Seed:  "); Serial.print( randSeed); 
    }

    mcRNum = random(100000000);  // used 8 digit value for PRIG selection of MC words
    pinx=0;
    postAscii=false;
    farns = deFF;
    echoChar=true;
    func = brightOne;
    clearDisp();
    while(Btn) {refreshWait(10); scanBtns();} // wait for button to clear

    Serial.print("\n FUNC: "); Serial.println(func);
    switch (func) {
      case 0:
        //btnReporting();    // exercises the buttons and Leds, and reports via Serial
        manageRate();   // Select / calibration Keying speed
        func=1;
        break;
      case 1:
        //debugTest();
        freeForm(); // practice Keying in what ever you want (BTW: .-.- will give you a C.R. )
        break;

      // --------------- Practice keying in all the different chacter types -------------------------
      case 2:
        info_ln("      Relay Psuedo Words. ");
        //                   ============
        Adv_Relay_exercise(true); // done with ditdah (&char) prompts
        break;
      case 3: // without ditdah prompts, only chars on screen
//        Serial.println("      Echo psuedo Words on Screen. ");
        Adv_Relay_exercise(false);
        break;

      // Listen to hundreds of random common Words
      //                              ============
      case 4:
        //algoChkSum(); // to verify that the Miller Shuffle Algo was ported correctly (sb: 6374753)
        Listen_Rand100s();   // long listening sessions of random common words 
        break;
      case 5:   // practice coping and tx Real Words from a large vocabulary
        MC_RxTx_Practice();  // mode 1&2: recieve only  mode 3&4: rcv and send
        break;

      //       	Test and advance you proficiency
      // ---------------------------------------
      case 6:
        MC_Lic_test();  //  Take a mock Morse Code Licensing Tests (Novice/General)
        break;
      case 7:
        echoChar=false;
        do {Tx_Line_test();}   // Transmit Line of 10+ chars then on all correct does 20+, then 30, 40, 50 ...
        while (!ESC);
        break;

      default:
        buZZ(1000);
    }
    ESC = false;
    Serial.println();
  }
}

void printDit(char mChar) {
    Serial.print("   Morse Char: ");  Serial.print(mChar);
    Serial.print("  0x"); Serial.print(mChar, HEX);
    Serial.print("    ave Period: ");  Serial.println(2*ditPeriod);
    Serial.print(" Traking Period: ");  Serial.println(pAveraging/4);
}

//================== major project funtional building blocks ===================

byte fetchMorseChar(bool echo) {
  if (oneKey) return(fetchMorse_1Key(echo));  // uses One button keying, based on duration of presses
//  else        return(fetchMorse_2Key(echo));  // uses basic Two button keying
  else        return(fetchMorse_2K_Auto(echo));  // uses Two button with fixed timing & Auto keying 
//  else        return(fetchMorse_22Key(echo));  // uses basic Two button keying improved p timing
}

// ---------- Morse code support functions ------------
byte fetchMorse_1Key(bool echo) { // get one char (ditdah set) using single button keying (straight CW key)
  byte mChar=0;
  int  p=0, ditPs;
  unsigned long pt;
  byte dit;
  //bool toggle=false;
  //static bool diTurn = true;
  //int lastavgDah;

  // update timing used prior to each whole character
  pAveraging=(avgDit+(avgDah/3))/2;  // avgDit & avgDah are 8x, so pAveraging is 4x
  //pAveraging=(avgDit+avgDah)/4;

  avgPeriod = pAveraging/4; // the running average maintains 4*resolution desired
  ditPeriod=avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   

  while(!(btn1 || btn3 || btn4)) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1) {
    p=0;
    pt = millis();  // time the period
    while (btn1) {
      dit_dah(2);   // output audio while key pressed (done for 2, sounds better than for 1)
      scanBtns();
    }
    p = millis()-pt;
    if (p>(5*ditPeriod)) p=5*ditPeriod;  // curb extream dah  (sb: 4 or 5)
    
    clearDisp();
//    dit = (p<avgPeriod);
//    dit = (p<(9*avgPeriod/10));  // try 1.8 ditPeriod the break point between a dit & a dah
    dit = (p<(7*avgPeriod/8)); // 1.75 of a dit
    // 1.8 is +80% of a dit, and -40% of a dah
    // 1.67 if +67% of a dit, and -45% of a dah
    ditPs=(p+ditPeriod/2)/ditPeriod;
    //lastavgDah=avgDah;

    if (p>(avgPeriod/8)) { // filter out spurious spikes
      if (mChar==0) { // setup first bit
        mChar = (dit)? 0x01 : 0xFE;
      } else {    // shift in other bits
        mChar = (mChar<<1) | dit;
      }
      // add into the running average 'pAveraging' (but not ones from left field)
      if (dit && (2*p>ditPeriod)) { // removed diTurn &&
        if (p>(ditPeriod+ditPeriod/2)) p=(ditPeriod+ditPeriod/2); // temper increases to avgDit
        pAveraging = (7*pAveraging/8) + p;
        avgDit = (7*avgDit/8) + p;
        //diTurn = !diTurn; // average in = # of dits & dahs
      }
  ///    if (!dit && p<(5*ditPeriod)) { // removed !diTurn &&
      if (!dit) { // removed !diTurn &&
        if (p>(4*ditPeriod)) p=(4*ditPeriod);
        if (p<(2*ditPeriod)) p=(2*ditPeriod); // temper decrease to avgDah
        pAveraging = (7*pAveraging/8) + (p/3);
        avgDah = (7*avgDah/8) + p;
        //diTurn = !diTurn;
      }
      //if (ditPs>4) avgDah=lastavgDah;  // dah too extream
    }
    else  {
      Serial.print("'");        // mark the rejected spurious spike with ' 
      if (enbBS) Serial.print("\b");  // and a BS (on supported monitors)
    }

    p=0;
    while (Btn==0) {  // wait for additional dit-dahs (or other)
      scanBtns();
      delay(1);
// the following needs to balance between an intervle long for inter dit-dah (1p) and short for inter char (3p)
//      if (++p>(7*ditPeriod/3)) {  // require only 2.3 (3 expected be the average)
//      if (++p>(2*ditPeriod)) {
      if (++p>(11*ditPeriod/5)) {  // require only 2.2 (3 expected be the average),  make "."  .-.-.- easier
        break;          // done with this char
      }
    }
  }

  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  if (func==1 && ditPs>5 && dida2Ascii(mChar)=='T' && enbBS) {  // long T used to delete prior character
    mChar=0; 
    Serial.print(char(0x08));Serial.print(' ');Serial.print(char(0x08));  // long T used to delete prior character
  } else if (echoChar) Serial.print(dida2Ascii(mChar));  // always output the alpha character (patched)
 
  debugDit(mChar);
  return (mChar);
}

// -----------------------   
byte fetchMorse_2Key(bool echo) { // using Two button keying (or a dual lever paddle)
     // not throurghly valided, may need adjustments to period timing
  //  static short int ditPeriod;
  byte mChar=0;
  int  p=0, p2=0, pCnt=0;
  bool dit;
  byte msn;
  bool toggle=false;
  byte ditdaCnt=0;

  avgPeriod = pAveraging/4;   // the running average maintains 4*resolution desired
  ditPeriod = avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   
  
  while(!Btn) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1 || btn2) {
    ditdaCnt++;
    p+=p2;
    dit = btn1;
    while (Btn) {
      scanBtns();
        if (btn1) {   // produce sound while pressing
          brightOne=7;
          ditAudio(1);
        } else {
          brightOne=7; litLED=8;
          dahAudio(2);
        }
      toggle = !toggle;
      p++;
      clearDisp();
    }

    if (mChar==0) { // setup first bit
      mChar = (dit)? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar<<1) | dit;
    }
    // count it into the running average 'pAveraging'
    if (dit) pCnt+=2;
    else pCnt+=3;   // compromise, as it takes 2 to key but 4 to reproduce
    
    p2=0;
    while (Btn==0) {  // wait for additional dit-dahs
      scanBtns();
      delay(1);
      if (++p2>(2*ditPeriod)) {
        break;      // done here, end of a char
      }
    }
  }

  if (ditdaCnt>1) {
    ditPeriod = p/pCnt;
  //    pAveraging = (3*pAveraging/4) + 2*(ditPeriod);  // provide tracking of avg. period of a ditdah
    pAveraging = (3*pAveraging/4) + 3*(ditPeriod);  // partially arbitarry & empirical
  }
  
  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  if (echoChar) Serial.print(dida2Ascii(mChar));  // always output the alpha character

  debugDit(mChar);
  return (mChar);
}

// -----------------------   
byte fetchMorse_22Key(bool echo) { // using Two button keying (or a dual lever paddle)
     // dit & dah differ in sound, but duration is simply based on button press duration.
     // TODO: Not throurghly valided, may need/want adjustments to period timing
  //  static short int ditPeriod;
  byte mChar=0;
  int  p=0, p2=0, pCnt=0;
  bool dit;
  byte msn;
  bool toggle=false;
  byte ditdaCnt=0;

  avgPeriod = pAveraging/4;   // the running average maintains 4*resolution desired
  ditPeriod = avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   
  
  while(!Btn) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1 || btn2) {
    ditdaCnt++;
    p+=p2;
    dit = btn1;
    t0=millis();
    while (Btn) {
      scanBtns();
        if (btn1) {   // produce sound while pressing
          brightOne=7;
          ditAudio(1);
        } else {
          brightOne=7; litLED=8;
          dahAudio(2);
        }
      toggle = !toggle;
      p++;
      clearDisp();
    }

    if (mChar==0) { // setup first bit
      mChar = (dit)? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar<<1) | dit;
    }
    // count it into the running average 'pAveraging'
    if (dit) pCnt+=2;
    else pCnt+=3;   // compromise, as it takes 2 to key but 4 to reproduce
    
    p2=0;
    while (Btn==0) {  // wait for additional dit-dahs
      scanBtns();
      delay(1);
      if (++p2>(2*ditPeriod)) {
        break;      // done here, end of a char
      }
    }
    if (dit) {  // just had a dit + an inter period
      pAveraging = (3*pAveraging/4) + (millis()-t0);  // provide tracking of avg. for 2p
      avgPeriod = pAveraging/4;
      ditPeriod = avgPeriod/2;
    }
  }
  
  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  if (echoChar) Serial.print(dida2Ascii(mChar));  // always output the alpha character

  debugDit(mChar);
  return (mChar);
}

// -----------------------   
byte fetchMorse_2K_Auto(bool echo) { // using Two button keying (or a dual lever paddle) with Auto repeat
     // not throurghly valided, may need adjustments to period timing
  static short int ditPeriod;
  byte mChar=0;
  int p2=0;
  int didaP;
  bool dit;
  byte msn;
  bool toggle=false;
  byte ditdaCnt=0;
  unsigned long t0;

  avgPeriod = pAveraging/4;   // the running average maintains 4*resolution desired
  ditPeriod = avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   
  // do not, do the moving averaging with auto keying
  
  while(!Btn) {scanBtns(); refreshWait(1);} // wait for button press
  while ((btn1 || btn2)) {
    ditdaCnt++;
  //    dit = btn1;
    if (btn1 && !btn2) dit=true;
    else if(btn2 && !btn1) dit=false;
    else dit= !dit;  // both are pressed

    didaP=(dit)? ditPeriod : 3*ditPeriod;
    t0=millis();
    while (didaP>millis()-t0) {
      scanBtns();
        if (dit) {   // produce sound while pressing
          brightOne=7;
          ditAudio(1);
        } else {
          brightOne=7; litLED=8;
          dahAudio(2);
        }
      clearDisp();
    }

    if (mChar==0) { // setup first bit
      mChar = (dit)? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar<<1) | dit;
    }
    
    delay(ditPeriod); // wait 1P standard time period between di-das
    p2=ditPeriod;
    stableBtns();
    while (Btn==0) {  // wait for additional dit-dahs
      scanBtns();
      delay(1);
      if (++p2>(2*ditPeriod)) {
        break;      // done here, end of a char
      }
    }
  }
  
  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  if (echoChar) Serial.print(dida2Ascii(mChar));  // always output the alpha character

  debugDit(mChar);
  return (mChar);
}

// -----------------------
bool isPuntc(byte charCode) { // identifies punctuation as we can be tolerant of delays keying them in (added 0-9 as well)
//  return (charCode>32 && !(toupper(charCode)>64 && toupper(charCode)<91) && !(toupper(charCode)>47 && toupper(charCode)<58));
  return (charCode>32 && !(toupper(charCode)>64 && toupper(charCode)<91));
}

//void debugTest() {
//}

// -----------------------   
char dida2Ascii(byte ditdahs) {
  byte ncode, ascii, msb, ndidas, i;
  const char nc2A[]  = "~EAIWRUSJP\nL~FVH~~QZYCXBOGKDMNT~";
//  const char dida5[] = "09 8   7 (   /=61    + &2   3 45";
//  const char dida6[] = " '   @ :  . ,\"    )?! $       - ";
  const char dida5[] = "09~8~~~7~(~~~/=61~~~~+~&2~~~3~45";
  const char dida6[] = "~'~~~@~:~~.~,\"~~~~)?!;$~~~~~~~-~";
  static bool lastP=false; // last char processed was a () type  

  if (ditdahs==0) return(0);

  // strip off all most significant extra 1s bits
  i=1;
  msb = (ditdahs & 0x80);
  while ((ditdahs<<i & 0x80) == msb) { i++;}
  ndidas = 8-i;
  ncode  = ditdahs & 0x1F;
  // decode according to its length
  if (ndidas<5)       ascii = nc2A[ncode];
  else if (ndidas==5) ascii = dida5[ncode];
  else if (ndidas>5)  ascii = dida6[ncode];

  if (!lastP) {
    if (ascii==')' && pOpen) ascii='(';
    if (ascii=='(') pOpen=false;
    //was    if (ascii==')' || ascii=="\n") pOpen=true;
    if (ascii == ')' || ascii == 0x0A) pOpen = true;
  }
  if (ascii=='(' || ascii==')') lastP=true;
  else lastP=false;

  //Serial.print(" ditdah code: ");  Serial.println(ditdahs);
  //Serial.print(" ascii code: ");  Serial.println(int(ascii));
  return (ascii);
}

// -----------------------   
byte Ascii2dida(char Achar) {
  byte ncode, i;
  const char nc2A[] = "~EAIWRUSJP~L~FVH~~QZYCXBOGKDMNT~~";
  int alen;

  Achar = toupper(Achar);
  //  move i-index past padding bits
  for (i=1; i<32; i++) {if (nc2A[i]==Achar) break;}
  if (i<32) {
    ncode = (i<0b10000) ? i : i+0b11100000;
  } else {  // handle #s, punc. & spec. chars
    if (Achar=='[' || Achar=='{')  Achar='(';
    if (Achar==']' || Achar=='}')  Achar=')';
    alen = sizeof(nonAlpha);  // sizeof(char)=1
    for (i=0; i<alen; i++) {if (nonAlpha[i]==Achar) break;}
    ncode = nonAlpha_dida[i];
  }
  return (ncode);
}

// --------------------------------------------   
void set_didaStr(byte ditdahs, char ddStr[]) {  // string returned in global var: 'didaStr'
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  byte i=1;
  byte bitState;

  if (ditdahs==0) return;

  while ((ditdahs<<i & 0x80) == (ditdahs & 0x80)) i++;
  int c=0;
  for (int b=i; b<8; b++) {
    if (ditdahs<<b & 0x80) {  // this bit is a 'dit'
      ddStr[c++]='.';    
    } else {  // its a dah
      ddStr[c++]='-';    
    }
  }
  ddStr[c++]=0;
}

//================= Main Morse Code Activities / Applets =================

// -------------
void freeForm() {  //  -------- allow free form Morse code practice until btn4 pressed
  byte ditdahs=0;
  char chrs=0;
  int p, ms;
  int innerWP;
  float wpm;

  easy=true;
  //  Serial.println(" Free Form   try:\nET MAIN SWORD GUK QYP FHJL ZXCVB -or- The five boxing wizards jump quickly (.-.-)");
  while (!btn4) {
    if (btn1 || btn2) {
      ditdahs = fetchMorseChar(0);  //echo);   // get dits & dahs appearing to form the MC of a character

      if (ditdahs!=0 && 0x0A!=dida2Ascii(ditdahs)) {chrs++; p=0;}
      litLED=(chrs-1)%8+1;
    }
    if (btn3) {  // tap B3: faster,  short hold B3: reduce speed
      ms=0;
      while (btn3) {
        scanBtns();
        refreshWait(1);
        ms++;
      }
      if (ms<200) ditPeriod = (6 * ditPeriod + 3) / 7; // pickup the pace 16+% faster
      else        ditPeriod = (7 * ditPeriod + 2) / 6; // slow down the pace 16+% slower
      wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
      Serial.print(" Keying rate: "); Serial.println(wpm);
      playStr((char*)"R", 0);  // .-.  2p for each
    }

    // ---------- checking for end of word
    if (ditdahs!=0) {
      innerWP=(easy)? 10 : 5;
      if (p<(innerWP*ditPeriod)) {if (++p==(innerWP*ditPeriod)) Serial.print(" ");}  // give 3+5 more than 7p
    }
    scanBtns();
    refreshWait(1);
  }
  reportRate();
}

// -----------------------   
void manageRate() {
  byte dida;
  byte ditdahs;
  //  bool echo=true;
  unsigned long t;
  int tmin, tmax, tsum;
  float wpm;
  static byte opt=2;
  byte chrs, newRate=0;
  bool fini=false;
  int ditP;

  while (btnDown()) ;
  clearDisp();
  while(!(btn4||fini)) {
    refreshWait(1);
    if ((msCnt%1750)<250) set_LEDs(0x55,1);  // slowly give short dim of 4 LEDs
    else set_LEDs(0x55,0);
    if (btnPressed()) {    // Button has been pushed
     clearDisp();
     refreshWait(1); 

      if (btn1) {  // ------------------------- 
        //  -------- first directly set ditdah Keying rate
  ///        setting=true;
        while (1) {
          ditP=ditPeriod;
          flashLed=1;
          refreshWait(10);
          do{ scanBtns();     // prompt with a flashing green led
            refreshWait(1);  
          } while (!BtnPressed);
          flashLed=0;
          if (btn1) {
            tmin=9999; tmax=0;
            tsum=0;
            for (int i=1; i<=7; i++) {
              t0=millis();
              do { 
                if (btn1) dit_dah(1);
                else delay(1);
                scanBtns();
              } while (!btn1Changed);
              t=millis()-t0;
              if (t>tmax) tmax=t;
              if (t<tmin) tmin=t;
              tsum+=t;
            }
            
            delay(100);
            if (2*tmin>(tsum-tmin)/6  &&  (tmax/2)<(tsum-tmax)/6) {
              ditPeriod = ((tsum-tmin)-tmax)/5;
  ///              setting=false;
            }
            else { buZZ(30); }
  ///            break;
          }
          else if(btn2) { ditPeriod=ditPeriod+ditPeriod/10; }
          else if(btn3) { ditPeriod=ditPeriod-ditPeriod/10; }
          else if(btn4) { break; }
          
          if(ditP!=ditPeriod) {
            Serial.print(" Average Period: "); Serial.print(ditPeriod);
            Serial.print("   WPM: "); Serial.println((1000.0*60.0) / (240.0*ditPeriod/5.0));
            //Serial.print(" min: "); Serial.print(tmin);  Serial.print(" max: "); Serial.println(tmax);         
            delay(500);
            pAveraging=8*ditPeriod;
            playStr((char*)"R",0);
  ///          break;
          }
        }
      }
      
      else if (btn2) { // --------  select a speed   8-36 WPM
        //   #s adjusted to compensate for processing overhead
        // set dit period to get WPM per Opt 1:10 2:12 3:14 4:16 5:18 6:20 7:22 8:24 ... 12:32
//TBD stack effect        Serial.println(" Select a Keying rate: 1)10 2)12 3)14 4)16 5)18 6)20 7)22 8)24   Audio Out will be faster."); 
        opt=get_Option(12, opt);   // ------ get speed Option
        if (ESC || opt==0) return;
        ditPeriod=620/(opt+4);   // empirically devined, for 10,12,14 ... 24
        //Serial.println(" Use Farswoth spacinng? [Yes/No] "); 
        //if (YesNo()) farns=deFF;
        //else farns=1;
        Serial.println(" Spacinng? [1:No  2-4: Farns:2x,3x,4x] ");
        farns=get_Mode(4);
      }
      else if (btn3) {  // --------  Save ditdah rate and Farnsworth factor
        savedPeriod = ditPeriod;
        EEPROM.write(eeSaveAddr, savedPeriod);
        savedFF = farns;
        EEPROM.write(eeSaveAddr+sizeof(savedPeriod), savedFF);
        wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
        debug_(" Saving Keying rate: "); debug_(wpm);
        debug_(" Farnsworth xfactor: "); debug_ln(savedFF);
        boop(150); beep(50);
      }
      else if (btn4) {
        fini=true;
        while (btnDown()) ;  // wait for release
      }
      pAveraging=8*ditPeriod;
      if (newRate!=ditPeriod) reportRate();
      newRate=ditPeriod;
      stableBtns(); // note due to debounce button states are not reported as changed with only one scan
    }
  }
}

void reportRate() {
  float wpm;
        wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
        //Serial.print(" avg. Period: "); Serial.println(pAveraging/4);
        Serial.print("\n Keying rate (in WPM): "); Serial.print(wpm);
        if (farns==1) Serial.println();
        else Serial.println("  + Farnsworth spacing");
        showLevel(wpm+0.5);
        refreshWait(2000);
        clearDisp();
}

// -------------
void Listen_Rand100s() {  // Listen to all of first top 100,200,500 or all; in random order
                         // You'll see the letters on the screen after the MC is sounded
  byte mode;
 ///  char wrd[21];
  byte wlen;
  int i, ri, wi; // ri:rand inx
  float wpm;
  int ditPace;
  int minInx, maxInx;
  byte llen=0;
  static int opt=1;
  int overall;

  Serial.println("    Listen to random words.");
  postAscii=true;
  // Modes:  1) first 100 words (now: any set of 100)    2) 1st 200    3) 1st 500   4) all words 0-899
  Serial.println(" select:   1) any set of 100   2) 1st 200    3) 1st 500   4) all words");
  mode = get_Mode(4); debug_ln(mode);
  if (mode == 0) return;
  Serial.print(" Mode #"); Serial.println(mode);

  minInx=0;
  if (mode==1) {  // for mode=1 allow to select any range of 100 eg: 400-500
    Serial.println("Select a set of 100 words. use <> & Btn1");
    opt=get_Option(numWrds/100+1,opt); // upto the 800-900 range
    maxInx = 100*opt;
    minInx=maxInx-100;
    if (opt==8) maxInx = numWrds;
  }
  else if (mode==2) maxInx = 200;  // can be anything 1-the total # of words (numWrds) you have 
  else if (mode==3) maxInx = 600; //500;
  else if (mode==4) maxInx = numWrds;  // 900

  ditPace = 5 * ditPeriod / 6; // set audio out higher than their keying rate
  //if (ditPace > 102) ditPace = 102; // maximum of 102 for minimum. 102 for 12 WPM (92 for 13+, 77 for 16 WPM, 69 for 18WPM)
  if (ditPace > 69) {ditPace = 69; farns=farns+2;} // if they want it slow go to 18wpm with added farnsworth
  
 //ditPace = 50; farns=1; // for fast debugging
  wpm = ((1000.0 * 60.0) / (240.0 * ditPace / 5.0));
  overall = 10*wpm * 50/(19*farns+31);   // overall WPM = keying WPM * 50/(19*farns+31)
  
  Serial.print("Char output rate: "); Serial.print(int(wpm+0.5));
  Serial.print(" WPM   Farns: "); Serial.print(farns); Serial.print("x");
  Serial.print("   overall WPM: "); Serial.println(overall/10.0);
  if (mode==2 || mode==3 || maxInx==100) {Serial.print("\n  Here is the top "); Serial.print(maxInx);}
  else if (mode==4) {Serial.print("\n  Here is all "); Serial.print(maxInx);}
  else {Serial.print("\n  Here is "); Serial.print(minInx); Serial.print("-"); Serial.print(maxInx-1);}
  Serial.println(" in random order:");
  
  clearDisp();
  refreshWait(1000);
  for (wi=0; wi<(maxInx-minInx); wi++) {
    dimFlash = 1 + (wi & 0x07);
    refreshDisp();

    ri = minInx +  shuffleIndex(wi, mcRNum, maxInx-minInx);
    // ---------------- get a word
    strcpy_P(wrd, (char *)pgm_read_word(&(words[ri])));
    wlen = strlen(wrd);
      
    // ---------------- MC ditdah out the word
    playStr(wrd, ditPace);
    stableBtns();  // scan for new button state
    if(btn4) break;
    else if (Btn) {             // go into Pause state
      set_LEDs(0x30, 3);
      while (Btn) {refreshWait(25); scanBtns();} 
      while (!BtnPressed) scanBtns(); // wait for next press to continue
      set_LEDs(0x30, 0);
      refreshWait(1500);   // give time to possibly return to paper & pencil
    }
    refreshWait(3 * ditPace * farns); // add 3+p more time to make up 7p between word period
    Serial.print(" ");
    llen=llen+wlen+1;
    if (llen>100) {llen=0; Serial.println();}
  }
  Serial.print("\n  Total word count was: "); Serial.println(wi);
}

// -----------------------   
void MC_RxTx_Practice() {
  byte opt, mode;
 ///  char wrd[22];
  byte ci, wlen, errs, lErrs, ccnt;
  bool errW;
  int i, ri, wi, w, p;
  byte ditdahs;
  char cinput,c;
  float wpm, wpmTkr;
  unsigned long t,t0;
  const int hiSpd=19;  // to use simplier words when working at lower speeds
  bool test=true; // normally its a prompted test, else its free form sending (CW keying)
  bool fini;
  int ditPace;

  // Modes:  1) RCV word at a time   2) RCV a series of words   3) Audio prompted SEND Practice
  //                                                            4) Screen promted only SEND
  // words are all randomly selected
  info_(" Mode #");
  mode=get_Mode(4); info_ln(mode);
  
  while (1) { // provide back to back operation, for long practice sessions
    if (farns>1) ditPace=ditPeriod;
    else ditPace=5*ditPeriod/6;    // set audio out higher than their keying rate
    if (ditPace>92) ditPace=92;  // maximum of 92 for minimum 13+ WPM (77 for 16 WPM)
    wpm=((1000.0*60.0) / (240.0*ditPace/5.0));
  
    if (mode<4) {
      Serial.print("\nChar output rate: "); Serial.print(int(wpm+0.5));
      Serial.print(" WPM   Farns: "); Serial.print(farns); Serial.println("x");
    }
    
    fini=false;
    ci=0;
    ccnt=0;
    errs=lErrs=0;
    w=0;
    wpmTkr=0;
    delay(1000);
    t0=(mode>2)? 0: millis();
    
    while (ccnt<60) {  // practice test is at leasr 60 chars
      pOpen=true;
      dimFlash= 1+ (w & 0x07);
      refreshWait(0);

      do {
        //wi = random(0,numWrds); // randomly select a 'word'
        wi = shuffleIndex(wInx++, mcRNum, numWrds); // randomly select a 'word' without stutering        
        strcpy_P(wrd, (char *)pgm_read_word(&(words[wi])));
 ///strcpy(wrd, ".;.");  // patch to repeatly only practice this string.
        wlen = strlen(wrd);
      } while ((wpm<hiSpd && wlen>6) || (wpm<4+hiSpd && wlen>8));
      // in the 900 element array there are 635 entries with 3-6 characters, 812 with 8 or less characters, 88 words more than 8
      if (mode<3) ccnt+=wlen;
  
      wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
      if (wpmTkr==0) wpmTkr=10*wpm;
      wpmTkr=(9*wpmTkr/10 + wpm);  // track & average over a practice round

      if (mode>2) {
        //Serial.print(" Keying WPM: "); Serial.print(wpm);
        //Serial.print("   WPM Traker: "); Serial.print(wpmTkr/10);
        if (errs!=lErrs) {Serial.print("    *** errs so far: "); Serial.print(errs);}
        lErrs=errs;
      }
      
      if (test) {
        if (mode<3) Serial.print(" "); else Serial.println();
        Serial.print(wrd);
      }
  
      for (i=0; i<strlen(wrd); i++) {
        c=wrd[i];     // use the same ditdah code for all brackets & braces
        if ( c=='[' || c=='{' || c=='(' || c=='}' || c==']' ) wrd[i]=')';
      }
      
      if (mode<3 && w==0) t0=millis();
      if (mode!=4) playStr(wrd,ditPace);
      if (mode==2) refreshWait(3*ditPace*farns);  // add 3+p more time to make up 7p between word period

      stableBtns();  // scan for new button state
      if(btn4) fini=true;
      else if (mode==2 && Btn) {             // go into Pause state
        set_LEDs(0x30, 3);
        while (Btn) {refreshWait(25); scanBtns();} 
        while (!BtnPressed) scanBtns(); // wait for next press to continue
        set_LEDs(0x30, 0);
        refreshWait(1500);   // give time to return to paper & pencil
      }
  
      if (mode==1 || mode==3) {    // handle one word at a time
        refreshWait(5*ditPace);  // delay in case btn held down
        while (!(btn1 || (btn2&&mode==3))) {   // ---------- wait to CW keying to begin
          refreshWait(1);
          scanBtns();
          if (btn4) {fini=true; break;}
          if (btn3Pressed || (btn2Pressed && mode==1)) {
            delay(1000);
            playStr(wrd,ditPace);    
          }
        }
      } else {
        scanBtns();
        if ((btn4 && mode==2) || ESC) return;
      }
  
      if (mode>2) {  // send (Xmit) modes
        // ---------------------------------  get the MC response
        for (ri=0; wrd[ri]!=0; ri++) {
          if (wrd[ri]==' ') continue;
          while(!Btn) {refreshWait(1); scanBtns();} // wait for button press
          dimFlash=litLED=0;
          refreshDisp();
          if (t0==0) t0=millis();
          ditdahs = fetchMorseChar(test);          
          litLED=1+ (w & 0x07);
          ccnt++;
          cinput = dida2Ascii(ditdahs);
          if (cinput=='(') cinput=')';  // for comparison only
          if (cinput!=toupper(wrd[ri])) { // -----------  did they get it right
            errs++;
            buZZ(50);
            delay(4*ditPeriod); // they may still be hitting the key
          }
          refreshWait(1);   
          p=0;
    
          if ( wrd[ri+1]==0) continue;
          errW = errs!=lErrs;  // was there already an err in this word
          while (!(btn1||btn2)) {  // ---------- wait for start of next letter (or end of word)
            if (btn4Pressed) {
              if (mode!=4) fini=true;
              if (test)  {test=false; Serial.println();}
              else {fini=true; break;}
            }
            refreshWait(1);
    //        if (++p>(5*ditPeriod)) { // give 3+5 more than 7p (3+4p) before thinking their done, as this is a test
            if (++p==(14*ditPeriod) && farns<3 && test && wrd[ri+1]!=' ' && !errW) {  //
              errs++;
              Serial.print(" __");
              buZZ(25);
 ///             ccnt++; // adj count so that WPM calc is not ill effected
             // wrd[ri+1]=0;
              // let then go on  //break;   // done here, go get next word
            }
            if (p==(4*ditPeriod) && wrd[ri+1]==' ') Serial.print("  ~ ");
            scanBtns();
          }
 //          if (p>(5*ditPeriod)) Serial.print("  ~ ");
          if (!test) ri=-1;
          if (fini) break;
        }
        if (fini) break;
      }
      if (fini) break;
      dimFlash=0;
      litLED=0; refreshDisp();
      if (mode==2) refreshWait(750);  // moment to get ready to write it down
      if (mode==3) {delay(30); beep(20);}  // going on to next word
      w++;
    }
   
    t=(millis()-t0);
    //Serial.print("\n Char cnt: "); Serial.print(ccnt); Serial.print("   time(msecs): "); Serial.print(t); Serial.print("   mins: "); Serial.print(t/60000.0);
    if (mode==1) t=t-(w*250); // -(250*11) for the input waits the code injected
    if (mode==2) t=t-(w*750); // -(750*11) for waits the code injected
    wpm=(ccnt/5.0)/(t/1000.0/60.0);    
    Serial.print("\n    over all ~WPM: "); Serial.println(int(wpm+0.5));
    if (mode<=2) {
      showLevel(wpm);
      refreshWait(2000); clearDisp();  
    } else {  // for modes >2
      {Serial.print("\n Current average Keying ditdah rate in WPM: "); Serial.print(wpmTkr/10);}
      showLevel((wpmTkr+5)/10);
      refreshWait(3000); clearDisp();
      if (test) {
        if (errs>0) {
          Serial.print("   Errors made: "); Serial.println(errs); 
          buZZ(150);
          showLevel(errs);
          refreshWait(3000); clearDisp();
        } else {
          Serial.println("    No errors were detected.");
        }
          Serial.print("  dah/dit Ratio (SB=3.0): "); Serial.println(float(avgDah)/avgDit);
      }
    }
    while (!btnPressed()) {delay(1);}  // wait for button press
    if (btn4 || ESC) return;  // B4->exit, else go again
    while(btnDown()) ;         // then wait for release
  }
}

char pickChar(byte type) {
  const char SpChar[] = "01234567890.,?/'-.,?/'-:\"+!=@$)&\n";  // don't change (but extras can be added for type=4)
  //const char ProbChar[] =   ".,'-CHWGVLYDCBXZSCHGVBCKVX?";  // using to focus on my weak spots, w/ 'type' = 5
  const char ProbChar[] =   "CCCHSBDLFGWPX=-ZQYKR";  // using to focus on my weak spots, edit as desired
  char c;
  byte  ci, mini, maxi, totMax;
  
  mini=1;  // default
  totMax=26+sizeof(SpChar)-1-1; // the sizof seems to be +1 (for a zero terminator)
  maxi=totMax;
  //  determin the char set to be used
  if      (type==1) {mini=1; maxi=26;}      // Alpha
  else if (type==2) {mini=1; maxi=40;}      // Alpha + Nums + punctation . thru +
  else if (type==3) {mini=27; maxi=57;}     // Nums + punctation . thru + (w/ maxi=51 for 57: . thru &)
  else if (type==4) {mini=37;}              // just punctation - special chars (includes .-.- newline, prompted for with "_")
  
//  Serial.print(" Maxi: "); Serial.println(maxi);

      if (type<5) {
        // get the character form the set he selected
        //ci = random(mini,maxi+1);  // multiplly repeats some chars while totally missing others 
        ci = shuffleIndex(pinx,mcRNum,2*(maxi-mini+1)) /2 + mini; // so all indexes occur 2 times in 2*possibles
        if (ci<27) {
          c = alpha[ci];
        } else {
          c = SpChar[ci-26];
        }
      } else {  // type 5: added for specific extra Problematic characters
        //ci = random(strlen(ProbChar)); // multiplly repeats some chars while totally missing others 
        maxi=strlen(ProbChar);
        ci = shuffleIndex(pinx,mcRNum,2*maxi) /2; // so all indexes occur 2 times in 2*possibles
        c = ProbChar[ci];
      }
      pinx++;
      return c;
}

void MC_Lic_test() {
  byte opt;
  char c;
  byte  w, wlen, numW, li, ci;
  char dida;
  int ditPace;
  byte ff;
//  float wpm;
  
  wlen=5;

  // took available stack RAM <200 & crashes: Serial.println("MC Test - 1:Nov. 5WPM   2:Gen. 12WPM  3&4:A-Z only    max: 2 errs");
  Serial.println(" 1:Nov. 5WPM   2:Gen. 12WPM  3&4:A-Z only");
  opt=get_Option(4,1);   // Novice 5 wpm       1: letters + .,$/   3: Novice letters only     2 errors allowed over 10 "words"
                         // General 12-13 wpm  2: letters + .,$/   4: General  "     "        2-5  "      "    over 25 "words"
                         // prior General MC tests were being done for 25WPM. & 5 errors were allowed 
  clearDisp();
  info_ln(opt);
  //if (ESC || opt==0) return;  // use opt==0 for problem letters
                      
  if (opt==0) {ditPace=ditPeriod; ff=farns+4;} 
  else if (opt==1 || opt==3) {ditPace = 82; ff=7;} // for Key@15WPM & 5WPM overall.  (102 for 12 WPM 92, for 13+, 82 for 15, 77 for 16 WPM)
  else {ditPace = 62; ff=2;}   // 12-13WPM overall keying at 20WPM.  (69 for 18WPM, 62 for 20 WPM)
//  wpm=((1000.0*60.0) / (240.0*ditPace/5.0)); 
//  Serial.print(" Audio out rate WPM: "); Serial.println(wpm);
  
  numW=(opt==2 || opt==4)? 25 : 10 ;
  do { 
    delay(1000);
    for (int w=1; w<=numW; w++) {
      for (li=0; li<wlen; li++)   { // generate random letters of a psuedo word
        if ((opt==1 || opt==2) && random(3)==1) c = pickChar(2);   // only 1/4 time allow #s.,/
        else  if (opt==0) c = pickChar(5);
        else c = pickChar(1);
        
        dida = Ascii2dida(c);
        if (c==0x0A) c='_';  // don't output \n

        playMorseCode(dida, ditPace);
        Serial.print(c);
        if (btn4) return;

        delay(2*ditPace*(ff));
      }
      Serial.print(" ");
      if (w%5==0) Serial.println();
      if (opt==0) delay(6*ditPace*(ff));  // praciticing, will lookup for a confim after each word
      else        delay(3*ditPace*(ff));
    }
//TBD   Serial.println("  Check your transcript.");  //  Nov: 5 Gen: 12 words in a row correct ");
    boop(50);
    while (!Btn) scanBtns();
  } while(btn1);
}


// -----------------------   
void Adv_Relay_exercise(bool audioPrompt) {
  char dd[10];  // stores ditdah given out
  byte ci, li, wlen, numW;
  int i,p,dl;
  static int opt=1;
  char dida;
  char c, cinput;
  byte correct, cwords;
  static int ditPace;
  static bool adjd=false;
  float wpm;
  bool lerr;

  //audioPrompt = (func==3);
  
  if (!adjd) ditPace = 5*ditPeriod/6;  // use 5/6 his current dit period, play at 1.2x keying speed
  // note: once manually overriden, new speed will get used
  wpm=((1000.0*60.0) / (240.0*ditPace/5.0)); 
// TBD reinstate  if (audioPrompt) {Serial.print(" Audio out rate to be used: "); Serial.println(wpm);}
  
  //Serial.println(" 1:A-Z  2:A-Z+#s+punct  3:#s+punct  4:punct+Spc.Chrs  5:Select Prob Chrs");
  Serial.print(" 1:A-Z 2:A+#+? 3:#+? 4:?+& 5:Select   ");
  opt=get_Option(5,opt);
  Serial.println(opt);
  if (ESC || opt==0) return;
                         // for testing transcribing, press Btn4 to output next "word"
  
  wlen = (opt<3)? 3:opt;
//  if (wpm>19) wlen=5;    // std length of "words" for testing is 5 chars
  if (wpm>19 && wlen<5) wlen++;
  numW = (opt==5)? 2:3;  // for opt5 we'll increament on each success
  if (wlen==3) numW=4;

  while(!btn4) {
    correct=0;
    cwords=0;
    clearDisp();
    while(btnDown()) refreshWait(1);
    delay(1200);

    for (int w=1; w<=numW; w++) {
      for (li=0; li<wlen; li++) { // generate random letters of a psuedo word
        c = pickChar(opt);
        dida = Ascii2dida(c);
        dd[li]=dida;
        if (c==0x0A) c='_';  // don't output \n
//        Serial.print(ci);
        Serial.print(c);

        if (audioPrompt) {  // prompt with ditdah audio and string
          playMorseCode(dida, ditPace);
          if (btn4) return;
          
          if (ditPace>120)  delay(2*2*ditPace);  // <10wpm use double std timing between char pause
          else delay(2*ditPace*(farns+4));  // give students +time (+farnsworth) to recallet what they just heard

          set_didaStr(dida, didaStr);
          Serial.print(" "); Serial.print(didaStr); Serial.print("    ");
        }
      }

      Serial.println();
      while (!(btn1 || btn2)) {  // wait for start of input
        refreshWait(1);
        scanBtns();
        if (btn4) return; // was: break; 
        if (btn3Pressed && audioPrompt) {
          delay(1000);           
          for (li=0; li<wlen; li++) {   // replay the 'word'
            playMorseCode(dd[li], ditPace);
            delay(2*ditPace*farns);
          }
        }
      }
      // do I realy need & want the functionality of the next code block ?
      // if (btn4) { // with btn4 break above ...
      //   while (btn4) { // handle ESC attempt
      //     delay(1);
      //     scanBtns();
      //     if (ESC) return;
      //   }
      //   delay(500);
      //   continue;  // simply goto next 'word'
      // }

      lerr=false;
      for (li=0; li<wlen; li++) { // get their reponse to the psuedo word given
        dida = fetchMorseChar(true);
        //if (btn4) return;
        Serial.print("  ");
        c=dida2Ascii(dd[li]);
        if (dida==dd[li] || c==0x0A) correct++;  // except anything for '_'
        else {lerr=true;buZZ(150,100);}
      }
      if (!lerr)  cwords++;
      else Serial.print("      Incorrect!");

      Serial.println();
      delay(1500);
    }
    
    if (opt<5) {
      Serial.print("\ncorrect: "); Serial.print(correct); 
      Serial.print("  out of ");  Serial.println(wlen*numW);
    } else {
      Serial.print("\nCorrect WORD count: "); Serial.print(cwords); 
      Serial.print("  out of ");  Serial.println(numW);
      if (cwords==numW) {
        numW+=1;
        Serial.print("  Very Good! Next try doing ");  Serial.println(numW);
      }
      correct=cwords;
    }
    if (correct==0) correct=1; // credit for participation
    showLevel(correct);
    refreshWait(2000); //clearDisp();
  }
}

// -----------------------   
void Tx_Line_test() {
  // btn 3 toggles Easy mode, where bad spacing is tollerated
  // each time you complete a line with no errors 10 characters is added to the test lines (upto 80 chars)
  ///  char wrd[21];
  char line[100];
  const char ending[] = ".,?!-/+.,?";
  byte ci, wc, wlen, wcnt, errs, ccnt, llen;
  int wi;
  int numWrds;
  byte i, xt;
  int p,  pspace;
  byte ditdahs;
  char cinput;
  float wpm, wpm_sum, wpmTkr;
  int wpm_min=0, wpm_max=0;
  byte chrs;
  bool lerr, ptime, goAgain=true;
  static byte mode=0;
  static bool easy=false;  // true= don't penalize poor spacing
  static byte maxC=10;
  unsigned long t, pauseT;
  char testC;
  int tv; // testing value

  //  for debuging ...
  //  unsigned long ranSeed = random(4000000000);
  //  randomSeed(ranSeed);
  //  Serial.print("\n randomSeed: "); Serial.println(ranSeed);

  if (mode==0) {
    Serial.println("    Tx Line Practice/Test\n");
 // TBD stack effect:    Serial.println("select 1: Words  2: w/ some #&...  3: full set  4: + random special");
    Serial.println("1: Words  2: + #&...  3: full set  4: + special");
    mode=get_Mode(4); Serial.println(mode);
  }
  if (mode==1) {numWrds=500; easy=true;}
  else if (mode==2) numWrds=800;
  else numWrds=sizeof(words)/sizeof(void *); // use accual size pointer for MCU inuse
  
 //  const int hiSpd=11;
 //  const byte hiSpd=18;  // to use simplier words when working at lower speeds
    const byte hiSpd=10;

  wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0));
 //  Serial.print("Starting char WPM rate: "); Serial.println(wpm);
  delay(250);  // avoid button bounce on entry
  
    i=0; ci=0;
    ccnt=0;
    wcnt=0;
    chrs=0;
    pauseT=0;
    while (ccnt<maxC) {    // predetermine 'word's for line totaling <80 chars
        //wi = random(0,numWrds); // randomly select a 'word'
        wi = shuffleIndex(wInx++, mcRNum, numWrds); // randomly select a 'word' without studdering
        strcpy_P(wrd, (char *)pgm_read_word(&(words[wi])));

  //#ifdef RP2040
  //     strcpy(wrd, (const char *) words[wi]);
 // #else
//       strcpy_P(wrd, (char *)pgm_read_word(&(words[wi])) );
 // #endif


    wlen = strlen(wrd);
    //while ((wpm<hiSpd && wlen>6) || (wpm<hiSpd+4 && wlen>8)); // limit word complexity as appropriate
    // in the 900 element array there are 635 entries with 3-6 characters, 812 with 8 or less characters, 88 words more than 8
    for (i=0; i<wlen; i++) {
      line[ci+i] = wrd[i];  // crude string copy
      if (mode==4 && random(3)==1) {   // 1/3 the time replace with a Special #s/chars
        wc = nonAlpha[random(1,strlen(nonAlpha)-1)];
        if (wc!=0x0A) line[ci+i] = wc;  // don't take '\n'
      }
    } 
    ci=ci+wlen;
    line[ci++]=' ';
    ccnt+=wlen;
    wcnt++;
  }
  if (mode>1) line[ci-1]=ending[random(0,strlen(ending))]; // ensure minimum practice of special chars
  else line[ci-1]=0x00; // too stuff do this instead
  
  line[ci]=0x00;  // mark end of string
  llen=strlen(line);
  
  do {   // allow single Btn4 press to restart a line (mostly for testing)
    while (Btn) {scanBtns(); delay(1); if (ESC) {mode=0; return;} }
    Serial.print("\n "); Serial.println(line); Serial.print(" ");

    chrs=0;
    pauseT=0;
    ci=0;
    wi=0;
    wpm_sum=0.0;
    wpmTkr=0;
    ccnt=0;
    t0=0;
    errs=0;
    xt=0;
    lerr=false;
    
    ptime=true; // allow time to pass before keying begins
    while (ci<llen) {
      cinput='#';
      p=0;
      while (cinput=='#') {  // ---------- get next MC character from user
        //dimFlash = 1+ (wi & 0x07); // used to show word progress, now we show ./- ratio
        litLED=(2*avgDah+(avgDit/2))/avgDit-3;  // leds = 2 2.5 3 3.5 4 ... Desire=3.0
        if (litLED>8) litLED=8;
        refreshWait(1);
        if (btnPressed()) {
          if (btn1 || btn2) {   // ---- they're keying in MC
            if (t0==0) t0=millis();
            ditdahs = fetchMorseChar(false);   // get dits & dahs appearing to form the MC of a character
            t1 = millis();
            pauseT=millis();
            cinput = dida2Ascii(ditdahs);
            if (easy && line[ci] == ' ' && (p > pspace/3)) {
              ci++;
              Serial.print(' ');
            }
          }
          if (btn3 && !easy) {easy=true; boop(20);}
          if (btn4) {
            if (ci>0) {goAgain=true; break;}
            // if (goAgain) btn4=0;  // go to next line
            // else mode=0; // reset operations
            if (!goAgain) {mode=0; ESC=true;}
            debug_("\n  Last ditDah:  ");
            set_didaStr(ditdahs, didaStr); debug_(didaStr);
            debug_("\n  ditPeriod= "); debug_(ditPeriod);
            debug_("  avgDit= "); debug_(avgDit/8);
            debug_("  avgDah= "); debug_ln(avgDah/8);
            return;
          } else goAgain=false;

        } else {    // ---- space between words detection
          delay(1);
          if (!ptime) {
            p++;
    //        if (p>(4*ditPeriod)) {  // time for a space between words (3p after last char + 4p = 7p)
  //          pspace = (line[ci]==' ')? 3 : 5;  // adj. to tolarate speed variations, at least
            pspace = (line[ci]==' ')? 5-2 : 3+xt;  // adj. to tolarate speed variations, at least
                                                    // xt normally =0 is for eXtra Time
            if ((!enbBS) && (line[ci]==' ')) pspace--; // don't forse so much time between words when they have no feedback prompt
            tv = pspace;
            pspace =  (pspace)*ditPeriod;  
            if (p > pspace) {  // if times up make their input a space
 //             Serial.print (" "); Serial.print (tv); 
              cinput =' ';  // if char expected it will err as too soon, if between words this must happen else err
            }
          }
        }
        
  //  OK now?    if (cinput==' ' && line[ci-1]==' ') cinput='#';  // allow extra space at begining of words
        // If last or next char is puncuation allow for a space in time.
  //      if (cinput==' ' && (isPunct(line[ci-1]) || isPunct(line[ci]))) cinput='#';
//        if (cinput==' ' && (isPunct(line[ci]) || isDigit(line[ci]))) cinput='#';   // also allow pause before #s
        // if (ESC || btn4) {
        //   if (ci>0) {goAgain=true; break;}
        //   if (goAgain) btn4=0;
        //   else mode=0; // reset operations
        //   debug_("\n  Last ditDah:  ");
        //   set_didaStr(ditdahs, didaStr); debug_(didaStr);
        //   debug_("\n  ditPeriod= "); debug_(ditPeriod);
        //   debug_("  avgDit= "); debug_(avgDit/8);
        //   debug_("  avgDah= "); debug_ln(avgDah/8);
        //   return;
        // } else goAgain=false;
      }
      xt=1; // sb: 0 to require good adherence to standard timing

      // debug: an error or something at times gets the input '({[' out of sync  (is mitigated)
      
      if (goAgain) break;
      // -----------  did we get the char we expected
      testC = line[ci];
      if (testC=='[' || testC=='{')  testC='(';
      if (testC==']' || testC=='}')  testC=')';
      if (cinput==')' && testC=='(')  cinput='(';  // mitigation, sync up open/close if out
      if (cinput=='(' && testC==')')  cinput=')';
      if (easy && line[ci]==' ' && cinput!=' ') {ci++; Serial.print('_');}
  //Serial.print("\ntestC="); Serial.print(int(testC));Serial.print("  :"); Serial.print(testC);Serial.println(":");
  //Serial.print("cinput="); Serial.print(int(cinput));Serial.print("  :"); Serial.print(cinput);Serial.println(":");
      // at time cinput of [ comes thru as ) & SB (   .. so patch ...
      if (cinput==toupper(testC)) {  // ---- got it as expected
        ci++;
        if (cinput!=' ') chrs++;
        ptime=(cinput==' ');
        if (isPunct(line[ci]) || isDigit(line[ci])) ptime=true;
        Serial.print(cinput);
        if (line[ci]==' ' || line[ci]==0x0) { // update WPM tracking after each word
          wi++;
          wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0));
          wpm_sum += wpm;
          if (wpmTkr==0) {wpmTkr=6*wpm; wpm_min=10*wpm; wpm_max=10*wpm;}
          wpmTkr=(5*wpmTkr/6 + wpm);  // track & average over a practice round
          //Serial.print("\n ditP WPM: "); Serial.print(wpm);
          //Serial.print("   Trkd WPM: "); Serial.print(wpmTkr/6);
          //Serial.print("   avg WPM: "); Serial.println(wpm_sum/wi);
          if ((10*wpm)<wpm_min) wpm_min=(wpm_min+10*wpm)/2; // track min. with a lit' temperment
          if ((10*wpm)>wpm_max) wpm_max=(wpm_max+10*wpm)/2;
        }
        if (lerr) xt=6; // if they are just righting a wrong allow x-time to get going again (was 3)
        lerr=false;
        
      } else {                       // ---- Not as expected
  //      easy=true;
        if (easy) {
          if (cinput==' ') continue;
          buZZ(25);
          if (lerr) continue;
          ci++;
          errs++;
          lerr=true; 
          Serial.print(char(tolower(cinput)));
          continue;
        }

        if (cinput==0x0A) cinput='~';  // don't echo \n, never correct & messes screen
        if (enbBS) {Serial.print(cinput); Serial.print(char(0x08));}
        
        //Serial.print (" "); Serial.print (tv); //Serial.print (" "); Serial.print (p);
        buZZ(25);
        if (!lerr) {  
          errs++;
          buZZ(75, 150);
        }
        lerr=true;
        ptime=true; // let them have time to regroup
  //      ptime=(line[ci]!=' ');  // if they gave wrong Letter-Symbol let them have time to regroup
        if (line[ci]==' ') {
          ci++;
          Serial.print("_");
          continue;
        }
      }
    }
  } while (goAgain);

  t=t1-t0;
  wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
//  Serial.print("\n current rate in WPM: "); Serial.print(wpm);
  // note: avg char code takes 12p, 4 char is 48p, 5* that is 240p, the std for 5WPM
//  Serial.print("  dah/dit Ratio (SB=3.0): "); Serial.print(float(avgDah)/avgDit);

  //Serial.print("\n Trending rate in WPM: "); Serial.print(wpmTkr/6);
  Serial.print("\n avg keying rate in WPM: "); Serial.print(wpm_sum/wi);
  //Serial.print("\nchrs: "); Serial.print(chrs); Serial.print("   t: "); Serial.println(t);

  wpm = (chrs/4.5)/(t/(1000.0*60.0)); // adj.d for the char distribution in our word databaase (w/#&...)
  if (wpm<1) wpm=1;
  if (maxC>10) {Serial.print("    over all ~ Words/Minute: "); Serial.println(wpm);}
  else Serial.println();
  int vard= int(100* (((wpm_max-wpm_min)/10.0)/wpm) );
  if (maxC<30 && vard==0) vard=1;
  Serial.print("           WPM varied: "); Serial.print(vard);
  Serial.print("%   dah/dit Ratio (SB=3.0): "); Serial.println(float(avgDah)/avgDit);
//  showLevel((wpmTkr+3)/6);
//  refreshWait(2000); clearDisp();
    delay(750);
  if (errs>0) {
    Serial.print(  "    # Incorrect characters:  "); Serial.println(errs); 
    buZZ(200+50*errs);
//    showLevel(errs);
//    refreshWait(1000); clearDisp();
  } else {
    Serial.println("   No Errors Detected!!");
    if (maxC<80) {
      maxC+=10;
      set_LEDs(0xFF,3); delay(250); // congratulations Flash
      set_LEDs(0xFF,0); delay(100);
      set_LEDs(0xFF,3); delay(250);     
      showLevel(maxC/10);
      //TBD reinstate: Serial.print("    Advance minimum line length to "); Serial.println(maxC);
      Serial.print("  New min. length "); Serial.println(maxC);
      refreshWait(2000); clearDisp();
    }
  }
}

void printTest_Chr_Ditdah() {
  const char SpChar[] = "01234567890.,?/'-:\"+!=@$)&;\n";
  byte ci, i, dida;
  char c;
  char hex[3];

//Serial.print(" nonAlpha 1 - "); Serial.print(sizeof(nonAlpha)-1);
//Serial.print("   nonAlpha_dida 1 -1 "); Serial.println(sizeof(nonAlpha_dida)-1);

  for (ci=1; ci<=52; ci++) {
    Serial.print(ci); Serial.print("  ");
    if (ci<27) {
      c = alpha[ci];
      Serial.print(c);
      dida = Ascii2dida(c);
    } else {
      i=ci-26;
      c = SpChar[i];
      Serial.print(c);
      //if (i>14) i-=4;  // adjust index to skip over repeated chars in SpChar[] 
      dida = nonAlpha_dida[i];
    }
    if (btnDown()) return;
    
    // check conversion utilities
    sprintf(hex, "%02X",Ascii2dida(c));
    Serial.print(dida2Ascii(Ascii2dida(c))); Serial.print(" : "); Serial.print(hex);

    playMorseCode(dida, 0);
    set_didaStr(dida, didaStr);
    Serial.print(" = ");
    Serial.print(didaStr); Serial.println("  ");
    delay(2*ditPeriod*farns);
    if (btnDown()) return;
  }
}

// ================== Local Support Utility Functions =================

// -----------------------   
void dit_dah(unsigned short int delayms){  // Used for ditdah audio while user has key pressed
                                          // higher freq & longer duty cycle for piezo
  delayms = delayms - (delayms/5);
  for (int i=0; i<delayms; i++) {
    digitalWrite ( AUDIO_OUT, HIGH); 
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(840);   // 750 840
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(560);   // 500 560
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
}

// -----------------------   
void ditAudio(unsigned short int delayms){  //  over duty cycle adjusted for freq.
                                  // ON duration extra needed for good piezo sound
//  delayms = delayms + (delayms/2);  // for use with 400/266
  delayms = delayms - (delayms/5);  // for use with 700/550
  for (int i=0; i<delayms; i++) {         // wait for 'delayms' millisecs
    digitalWrite ( AUDIO_OUT, HIGH);
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(840);    // tried: 700 600 400 456 best? 700 825 840
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(560);   // tried: 550 400 266 210 best? 500 550 560
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
}

// -----------------------   
void dahAudio(unsigned short int delayms){  // lower freq ( w/ longer ON duty cycle for Piezo)
  delayms = (2*delayms)/3;
  for (int i=0; i<delayms; i++) {         // wait for delayms ms
    digitalWrite ( AUDIO_OUT, HIGH); 
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(870);    // 900
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(580);    // 600
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
}

// -----------------------   
void playMorseCode(byte ditdahs, int pace) {
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  short int ditPeriod;
  byte i=1;
  byte bitState;

  ditPeriod = pAveraging/8;    // set according to their current pace
  ditPeriod = 8*ditPeriod/10;  // but 80% timing, for 1.25 the speed
  if (pace) ditPeriod=pace;    // unless passed in

  if (ditdahs==0) return;
  while ((ditdahs<<i & 0x80) == (ditdahs & 0x80)) i++;
  for (int b=i; b<8; b++) {
    if (ditdahs<<b & 0x80) {  // this bit is a 'dit'
      brightOne=7;
      ditAudio(ditPeriod);
      brightOne=0;
      refreshWait(ditPeriod);      
    } else {  // its a dah
      dim[7]=dim[8]=true;
      dahAudio(3*ditPeriod);
      dim[7]=dim[8]=false;
      refreshWait(ditPeriod);
    }
    scanBtns();  // needed to keep button state upto date (for btn4 esc)
  }
////  refreshWait(2*ditPeriod);
}

// -----------------------   
void playStr(char *wrd, int pace) {          // ment for messsage announcements
  float ff=farns;
  char c;
  byte ditdahs;
  int ditPace;
  byte dimF;
  
  dimF=dimFlash;
  dimFlash=0; // stop position flash in MC_Practice during this word output
  refreshWait(10);
  
  ditPace=ditPeriod;
  if (pace) ditPace=pace;    // unless passed in

  for (int i=0; i<strlen(wrd); i++) {
    ditdahs=Ascii2dida(wrd[i]);
    if (ditdahs==0) delay(4*ditPace*ff);  // handle timing for the likes of a space
    playMorseCode(ditdahs, ditPace);
    //Serial.print(wrd[i]); Serial.print(ditdahs);
    if (postAscii) Serial.print(wrd[i]);
//    refreshWait(3*ditPace*ff-ditPace);   // +2p for between character
    refreshWait(2*ditPace*ff);   // +2p for between character
    scanBtns();
//    if (btn4) return;
    if (Btn) return;
  }
//  refreshWait(3*ditPace*ff);  // add 3+p more time to make up 7p between word period
//  now caller's responsibility
    dimFlash=dimF;
}

// ================ Generally Useful Support Utility Functions =============== 
void showLevel(byte level) {
  int n;
  clearDisp();
  refreshWait(250);
  for (int i=0; i<level; i++) {
    beep(10);
    n = (i%nleds)+1;  // n wraps around as i exceeds nleds
    dim[n]=true;
    brightOne=n;
    refreshWait(100);
  }
}

// -----------------------   
byte get_Mode(byte modeCnt) {  // modeCnt is the max selectable mode #
  byte mode;
  byte ledSet=0;
  byte ledMask=0;
  int cylceTime=650;
  
  if (modeCnt>4) return 0;
  clearDisp();
  for (int i=1; i<=(2*modeCnt); i++) {dim[i]=true;} // set relavent leds dim
  while(btnDown()) ;  // wait for button release
  while(! Btn) {
    if (0==(msCnt%cylceTime)) {  // every 1/2 sec highlight a posible selection
      ledSet = (ledSet+1) % 4;  // cycles 0 1 2 3
      if (ledSet>=modeCnt) ledSet=0;
      ledMask = 0x03 << 2*(ledSet);
      set_LEDs(ledMask,2);
    } if (200==(msCnt%cylceTime)) {
      set_LEDs(ledMask,1);  // undo last lit
    }
    refreshWait(1);
    scanBtns();
  }
  set_LEDs(ledMask,1);  // undo last lit
  mode = btnNum;
  // highlight bright their selection
  ledMask = 0x03 << 2*(mode-1);
  set_LEDs(ledMask,3);
  while(Btn) {
    refreshWait(1);
    scanBtns();
    if (ESC) return(0);
  }
  clearDisp();
  return mode;
}

// -----------------------   
byte get_Option(byte optCnt, int dflt) {  // optCnt is the max # option, dflt is the default value
  int cnt;
  clearDisp();
  for (int i=1; i<=optCnt; i++) {dim[(i-1)%nleds+1]=true;}
  cnt=dflt;
  while(btnDown()) ;  // wait for button release
  while(! btn1) {
      flashLed=(cnt-1)%nleds+1;
      for (int i=1; i<=nleds; i++) {lit[i]=(i<cnt);}  // lite those up to cnt
      refreshWait(1);
      if (btnPressed()) {
        if (btn2Pressed) cnt = (cnt==0)?  optCnt : cnt-1;  // -- move left Allow 0:debugging
        if (btn3Pressed) cnt = (cnt%optCnt) + 1;              // ++ moves right
        //if (btn4Pressed) cnt = (cnt%optCnt) + 1; // allow most play with only 2 buttons (1&4)
      }
      if (ESC) return(0);
  }
  //clearDisp(); this will reset cnt !
  return(cnt);
}

bool YesNo() {
  clearDisp();
  set_LEDs(0xC3,1);  // dimly lite the Grn & Red LEDs
  while(btnDown()) ;  // wait for button release
  while(! (btn1 || btn4)) {
      refreshWait(1);
      if((millis()/500)%2) { //  toggle every 500 ms
          set_LEDs(0xC0,2);
          set_LEDs(0x03,1);
      } else {
          set_LEDs(0x03,2);
          set_LEDs(0xC0,1);
      }
      scanBtns();
  }
  clearDisp();
  return(btn1);
}

// --------------------------
// produce a requested number of randomly sorted indexes  ~= FisherYates Algo
void ranIndexes(byte rIndexes[], byte cnt) {  
  char wi[80];  // working set of indexes
  int i, indx;
  byte tmp;
  
  for (int c=1; c<=cnt; c++) wi[c]=c; // preset the collection of indexes
  
  for (int s=cnt; s>0; s--) {  // simple fast random sort
    i = random(1,s);
    rIndexes[s]=wi[i];
    wi[i]=wi[s];
  }
  // ensure there are not any consecutive entries, not for randomness but AB_exersice desired behavior
  for (i=1; i<cnt; i++) {
    if (rIndexes[i]+1==rIndexes[i+1]) {
      tmp = rIndexes[i];
      rIndexes[i]=rIndexes[i+1];
      rIndexes[i+1]=tmp;
    }
  }
}

// --------------------------------------------------------
// produce a shuffled Index given a base Index, a random seed and the length of the list being indexed
// for each inx: 0 to nlim-1, unique indexes are returned in a psuedo "random" order
// This is eXtra-lite implementation of the Miller Shuffle Algorithm, go see https://github.com/RondeSC/Miller_Shuffle_Algo
// This algorithm is also know as a PRIG function (Psuedo Random Index Generator)   chksum: 6174976
short int shuffleIndex(unsigned short inx, unsigned long mixID, short nlim) {
	const short p1 = 9949, p2 = 9973;  // limits nlim to < 9949
	short si, r1, r2, rx;

	mixID += 131 * (inx / nlim);  // have inx overflow effect the mix
	si = (inx + mixID) % nlim;    // cut the deck

	r1 = mixID % p2;	// set of fixed randomizing values
	r2 = mixID % p1;
	rx = (mixID / nlim) % nlim + 1;

	// perform conditional multi-faceted mathematical spin-mixing  (average # of si recalcs performed is ~1.5)
	if (si % 3 == 0) si = (((long)(si / 3) * p1 + r1) % ((nlim + 2) / 3)) * 3; // spin multiples of 3 
	if (si % 2 == 0) si = (((long)(si / 2) * p2 + r2) % ((nlim + 1) / 2)) * 2; // spin multiples of 2 
	if (si < rx) si = ((long)si * p2 + r1) % rx;                       // mix random half one way
	else         si = ((long)(si - rx) * p1 + r2) % (nlim - rx) + rx;  // and the other another

	return(si);  // return 'Shuffled' index
}

// -----------------------
unsigned long algoChkSum() {  // does a Simple Shifting Check Sum (both value and sequence dependant)
  unsigned int nlimit=256;
  unsigned long randCut;
  unsigned long item, csum, lim;
  char sh;
  
  randCut = 314159;   //
  csum=sh=item=0;
  for (lim=nlimit; lim<(nlimit+3); lim++) { // use 3 different range limits
	  for (int i=0; i<(2*lim); i++)		// better, so as to test i>nlimit
	  {
		  item = shuffleIndex(i, randCut, lim); 
		  csum += (item<<sh);
		  if (++sh==8) sh=0;
	  }
  }
  Serial.print("  Miller Shuffle Algorithm chkSum: "); Serial.println(csum);
  return (csum);
}

// -----------------------   
bool btnDown() {
  refreshWait(1);
  scanBtns();
  return (Btn!=0);
}

// -----------------------   
void stableBtns() { // get a stable state of the buttons post debouncing
  refreshWait(1);
  scanBtns();
  while(!btnStable) {
    refreshWait(1);
    scanBtns();
  }
}

void LOUDbeep(int delayms){  // works for both Buzzers & Speakers
  for (int i=0; i<delayms; i++) {         // wait for a delayms ms
    digitalWrite ( AUDIO_OUT, HIGH);
    digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(500);
    digitalWrite ( AUDIO_OUT, LOW);
    digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(500);
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  digitalWrite ( AUDIO_NEG, LOW);
}  

void LOUDbuZZ(int delayms){  // works for both Buzzers & Speakers
  for (int i=0; i<delayms; i+=5) {       // wait for >= delayms ms
    LOUDbeep(2);
    digitalWrite ( AUDIO_OUT, HIGH);
    digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(1500);
    digitalWrite ( AUDIO_OUT, LOW);
    digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(1500);
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  digitalWrite ( AUDIO_NEG, LOW);
}  
